/**
  ******************************************************************************
  * @file   : Firmware.java 
  * @author : Softweb Solutions An Avnet Company
  * @modify : 17-Mar-2021
  * @brief  : Firmware part for JAVA SDK v3.0
  ******************************************************************************
*/

/**
 * Hope you have installed the Java SDK v3.0 as guided in README.md file or from documentation portal. 
 * Import the IoTConnect SDK package and other required packages
 */

package iotconnect.sdk;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import iotconnect.sdk.dao.SDKOptions;
import iotconnect.sdk.dao.d2c.D2CResponse;
import iotconnect.sdk.service.SDKCallback;
import iotconnect.sdk.service.SDKClient;
import iotconnect.sdk.service.SDKClientImpl;
import iotconnect.sdk.service.TwinCallback;
import org.apache.commons.lang3.StringUtils;

import java.util.*;

public class Firmware {

    static SDKClient sdkClient = null;
    public static boolean isDeviceConnected = false;

    public static void main(String[] args) throws Exception {

        CBImplementor cb = new CBImplementor();
		/*
        ## Prerequisite parameter to run this sample code
        - cpId              :: It need to get from the IoTConnect platform "Settings->Key Vault".
        - uniqueId          :: Its device ID which register on IotConnect platform and also its status has Active and Acquired.
        - env               :: It need to get from the IoTConnect platform "Settings->Key Vault".
        - dataSendInterval  :: send data frequency in seconds
        - disconnectDuration:: Device connection close after defined time in seconds to call dispose() method. Keep "0" if don't need disconnect
                               the device.
        - sdkOptions        :: It helps to define the path of self signed and CA signed certificate as well as define the offline storage params
        */
        String cpId = "<< CPID >>";
        String env = "<< Environment >>"; //Environment DEV, QA, POC, AVNET, PROD - (Default - PROD)
        String uniqueId = "<< Device UniqueID >>";
        Integer dataSendInterval = 10; // send data frequency in seconds
        Integer disconnectDuration = 0; //Seconds // Keep "0" If no need to disconnect

        /** 
		* sdkOptions is optional. Mandatory for "certificate" X.509 device authentication type
		* "certificate" : It indicated to define the path of the certificate file. Mandatory for X.509/SSL device CA signed or self-signed authentication type only.
		* 	- SSLKeyPath: your device key
		* 	- SSLCertPath: your device certificate
		* 	- SSLCaPath : Root CA certificate
		* "offlineStorage" : Define the configuration related to the offline data storage 
		* 	- disabled : false = offline data storing, true = not storing offline data 
		* 	- availSpaceInMb : Define the file size of offline data which should be in (MB)
		* 	- fileCount : Number of files need to create for offline data
		* Note: sdkOptions is optional but mandatory for SSL/x509 device authentication type only. Define proper setting or leave it NULL. If you not provide the offline storage it will set the default settings as per defined above. It may harm your device by storing the large data. Once memory get full may chance to stop the execution.
		*/ 

        SDKOptions sdkOptions = new SDKOptions();
        sdkOptions.setDisabledOfflineStorage(false);    //default value = false, false = store data, true = not store data
        sdkOptions.setAvailSpaceInMb(2);               //in MB Default value = unlimited
        sdkOptions.setFileCount(5);                   //  Default value = 1
//      sdkOptions.setSSLCaPath("<<filepath>>/ca.cert.pem");
//      sdkOptions.setSSLCertPath("<<filepath>>/cert.pem");
//      sdkOptions.setSSLKeyPath("<<filepath>>/key.pem");

        /*
        Type    : Object Initialization "new SDKClient()"
        Usage   : To Initialize SDK and Device connection
        Output  : Callback methods for command and twin properties
        Input   : cpId, uniqueId, sdkOptions, env as explained above and deviceCallback and twinUpdateCallback is callback functions
        */
        if(!isDeviceConnected){
            sdkClient = new SDKClientImpl(cpId, uniqueId, cb, cb, sdkOptions, env);
        }else{
            System.out.println("Device is already connected");
        }

        // Non Gateway
        String deviceData = "[{'data':{},'time':'','uniqueId':'<<uniqueId>>'}]";

        // Gateway Device
        /*String deviceData = "[{'data':{},'time':'','uniqueId':'<<uniqueId>>'}," +
                "{'data':{},'time':'','uniqueId':'<<Child_UniqueID>>'}]";*/

		/**
		* Add your device attributes and respective value here as per standard format defined in sdk documentation
		* "time" : Date format should be as defined //"2021-01-24T10:06:17.857Z" 
		* "data" : JSON data type format // {"temperature": 15.55, "gyroscope" : { 'x' : -1.2 }}
		*/
        TimerTask sendDataTask = new TimerTask() {

            @Override
            public void run() {
                try {
                    // Device to Cloud data publish
                    sdkClient.sendData(deviceData);
                } catch (JsonProcessingException e) {
                    System.out.println("Error While Publishing Data : " + e.getMessage());
                }
            }
        };

        Timer sendDateTimer = new Timer();
        sendDateTimer.schedule(sendDataTask, 0, dataSendInterval * 1000);

        /*
        Type    : Public Method "getAllTwins()"
        Usage   : To get all the twin properties Desired and Reported
        Output  : All twin property will receive in callback function "twinUpdateCallback()"
        */
		//      sdkClient.getAllTwins();

		
		/*
         Type    : Public Method "getAttributes()"
         Usage   : Send request to get all the attributes
         Input   :
         Output  :
         */
        //        sdkClient.getAttributes();
		
        /*
        Type    : Public Method "dispose()"
        Usage   : Disconnect the device from cloud
        Note : It will disconnect the device after defined time
        */
        if (disconnectDuration > 0) {
            TimerTask disposeTask = new TimerTask() {

                @Override
                public void run() {
                    if(isDeviceConnected){
                        sdkClient.dispose();
                        isDeviceConnected = false;
                    }
                }
            };

            Timer disposeTimer = new Timer();
            disposeTimer.schedule(disposeTask, 0, disconnectDuration * 1000);
        }
//        sdkClient.close();
//        System.exit(0);
    }

    //Callback class
    static class CBImplementor implements SDKCallback, TwinCallback {

        /*
		Type    : Callback Function "deviceCallback()"
		Usage   : Firmware will receive commands from cloud. You can manage your business logic as per received command.
		Output  : Receive device command, firmware command and other device initialize error response
		Input   :
		*/
        @Override
        public void deviceCallback(D2CResponse d2CResponse) {
            try {
                if (d2CResponse != null && d2CResponse.getD2cData() != null) {
                    System.out.println("CallBack Data :: " + new ObjectMapper().writeValueAsString(d2CResponse));

                    if (d2CResponse.getCmdType().equals("0x01")) {        // for "0x01" device command
                        if (StringUtils.isNotBlank(d2CResponse.getD2cData().getAckId()) && StringUtils.isNotBlank(d2CResponse.getD2cData().getUniqueId())) {
                            /*
                            Type    : Public Method "sendAck()"
                            Usage   : Send device command received acknowledgment to cloud
                            - status Type
                                st = 6; // Device command Ack status OR
                                st = 4; // Failed Ack
                            - Message Type
                                msgType = 5; // for "0x01" device command
                            */
                            Integer msgType = 5;
                            String data = "[{'data': {'ackId': '" + d2CResponse.getD2cData().getAckId() + "','st': 6,'msg':'','childId': ''},'uniqueId':'" + d2CResponse.getD2cData().getUniqueId() + "'}]";
                            sdkClient.sendAck(data, msgType);
                        }

                    } else if (d2CResponse.getCmdType().equals("0x02")) {  // for "0x02" Firmware OTA Command
                        String childId = null;
                        List<Map<String, Object>> urlData = d2CResponse.getD2cData().getUrls();
                        if (urlData != null && urlData.size() > 0 && urlData.get(0).containsKey("uniqueId")) {
                            childId = (String) urlData.get(0).get("uniqueId");
                        }
                        if (StringUtils.isNotBlank(d2CResponse.getD2cData().getAckId()) && StringUtils.isNotBlank(d2CResponse.getD2cData().getUniqueId())) {
                             /*
                            Type    : Public Method "sendAck()"
                            Usage   : Send firmware command received acknowledgement to cloud
                            - status Type
                                st = 7; // firmware OTA command Ack status OR
                                st = 4; // Failed Ack
                            - Message Type
                                msgType = 11; // for "0x02" Firmware command
                            */
                            Integer msgType = 11;
                            String data = "[{'data': {'ackId': '" + d2CResponse.getD2cData().getAckId() + "','st': 7,'msg':'','childId': " + childId + "},'uniqueId':'" + d2CResponse.getD2cData().getUniqueId() + "'}]";
                            sdkClient.sendAck(data, msgType);
                        }
                    }else if(d2CResponse.getCmdType().equals("0x16") && StringUtils.isNotBlank(d2CResponse.getD2cData().getUniqueId())){ // 0x16 Device connection status command
                        if(d2CResponse.getD2cData().getCommand().equals("true")){
                            isDeviceConnected = true;
                            System.out.println("DeviceId ::: [" + d2CResponse.getD2cData().getUniqueId() + "] :: Device Connected :: " + new Date());
                        }else{
                            isDeviceConnected = false;
                            System.out.println("DeviceId ::: [" + d2CResponse.getD2cData().getUniqueId() + "] :: Device Disconnected :: " + new Date());
                        }
                    }
                }
            } catch (Exception e) {
                System.out.println("Error While Sending Acknowledgement to HUB" + e.getMessage());
            }
        }

        /*
        Type    : Callback Function "twinUpdateCallback()"
        Usage   : Manage twin properties as per business logic to update the twin reported property
        Output  : Receive twin properties Desired, Reported
        Input   :
        */
        @Override
        public void twinUpdateCallback(Map<String, Object> twinProperties) {
            // Send Twin property Reported Value
            if (twinProperties.size() > 0) {
                Map<String, Object> twinPropMap = new HashMap<>();
                try {
                    Iterator<Map.Entry<String, Object>> mainIterator = twinProperties.entrySet().iterator();
                    while (mainIterator.hasNext()) {
                        Map.Entry<String, Object> mainEntry = mainIterator.next();
                        if (mainEntry.getKey().equals("desired")) {
                            Map<String, String> lhmap1 = (Map<String, String>) mainEntry.getValue();
                            Iterator<Map.Entry<String, String>> finalIterator = lhmap1.entrySet().iterator();
                            while (finalIterator.hasNext()) {
                                Map.Entry<String, String> entry = finalIterator.next();
                                if (!entry.getKey().toString().equalsIgnoreCase("$version")) {
                                    twinPropMap.put(entry.getKey(), entry.getValue());
                                }
                            }
                        } else {
                            if (!mainEntry.getKey().toString().equalsIgnoreCase("$version") && !mainEntry.getKey().toString().equalsIgnoreCase("reported")) {
                                twinPropMap.put(mainEntry.getKey().toString(), mainEntry.getValue().toString());
                            }
                        }
                    }
                } catch (Exception e) {
                    System.out.println("Error while generating Twin Update Request " + e.getMessage());
                }
                try {
                    sdkClient.updateTwin(twinPropMap);
                    System.out.println("Twin Property Updated");
                } catch (Exception e) {
                    System.out.println("Error While Updating Twin Property " + e.getMessage());
                }
            }
        }
    }
}